import javax.swing.*;
import javax.swing.table.*;
import java.awt.event.*;
import java.awt.*;

import javax.swing.plaf.metal.*;

/**
 * A base class that adds functionality for sorting a Table by column, 
 * obtaining tooltips for given cells, creating popup menus for given cells/rows/cols,
 * and double-click actions for given cells/rows/cols
 *
 * @see DeluxeTableModel
 */
public class DeluxeTable extends JTable {

    /**
     * our deluxe tablemodel reference
     */
    private final DeluxeTableModel _deluxeModel;

    private int _maximumColumnWidth = 128;

    /**
     * set the maximum column width to be applied to all columns.
     * used as a sort of stopgap sanity-check to keep tables from getting out of hand
     * @param num the maximum width of all columns, in pixels
     */
    public void setMaximumColumnWidth( int num ) {
        _maximumColumnWidth = num;
    }
    
    /**
     * A DeluxeTable will only accept a TableModel of the Deluxe variety.
     */

    public DeluxeTable( DeluxeTableModel model ) {
        super( model );
        _deluxeModel = model;
        
        JTableHeader header = getTableHeader();

        //
        // listen for double clicks on the header, for table sorting
        //

        header.addMouseListener( new MouseAdapter() {
                        public final void mouseClicked( MouseEvent evt ) {
                                if ( evt.getClickCount() % 2 == 0 ) {
            
                                        TableColumnModel colModel = getColumnModel();
                                        int column = colModel.getColumnIndexAtX( evt.getX() );
                                        int col = convertColumnIndexToModel( column );
                                        evt.consume();
                                        _deluxeModel.sortOnColumn( col );
                                }
                        }
                } );

        //
        // listen for double clicks on the table, and popup menu events
        //

        addMouseListener( new MouseAdapter() {
                
                public void mousePressed( MouseEvent evt ) {
                    if ( evt.isPopupTrigger() ) {
                        handlePopupEvent( evt );
                    }
                }
                
                public void mouseReleased( MouseEvent evt ) {
                    if ( evt.isPopupTrigger() ) {
                        handlePopupEvent( evt );
                    }
                }

                public void mouseClicked( MouseEvent evt ) {
                    if ( evt.isPopupTrigger() ) {
                        handlePopupEvent( evt );
                    }
                    else if ( evt.getClickCount() > 1 && 
                         ( evt.getModifiers() & InputEvent.BUTTON1_MASK ) != 0 ) {
                        handleDoubleClickEvent( evt );
                    }
                }
            } );
        
        int rowheight = getFontMetrics( MetalLookAndFeel.getUserTextFont() ).getHeight();
        setRowHeight( rowheight );
    }
    
    /**
     * Implementation of getToolTipText to return tooltips
     */
    public final String getToolTipText( MouseEvent evt ) {
        TableColumnModel colModel = getColumnModel();
        int column = colModel.getColumnIndexAtX(evt.getX());
        int colIndex = convertColumnIndexToModel(column); 
        int padding = getIntercellSpacing().height;
        int rowHeight = getRowHeight();
        int rowIndex = (evt.getY() / (padding + rowHeight));

        if (rowIndex < 0 || rowIndex >= getRowCount() || colIndex < 0 )
            return null;
        
        return _deluxeModel.getToolTipAt( rowIndex, colIndex );
    }
    
    /**
     * subclasses should override this if they want popup menus.  This method
     * does nothing by default.
     */
    protected void doPopup( MouseEvent evt, int row, int col ) {}

    /**
     * subclasses should override this if they want double click actions.  This method
     * does nothing by default.
     */
    protected void doDoubleClick( MouseEvent evt, int row, int col ) {}

    /**
     * this method is called by the default mouse event listener on popup triggers
     */
    
    public final void handlePopupEvent( MouseEvent evt ) {

        TableColumnModel colModel = getColumnModel();
        int column = colModel.getColumnIndexAtX(evt.getX());
        int colIndex = convertColumnIndexToModel(column); 
        int padding = getIntercellSpacing().height;
        int rowHeight = getRowHeight();
        int rowIndex = (evt.getY() / (padding + rowHeight));

        if (rowIndex < 0 || rowIndex >= getRowCount() || colIndex < 0 )
            return;
        
        doPopup( evt, rowIndex, colIndex );
    }

    /**
     * this method is called by the default mouse event listener on double clicks
     */
    
    public final void handleDoubleClickEvent( MouseEvent evt ) {
        TableColumnModel colModel = getColumnModel();
        int column = colModel.getColumnIndexAtX(evt.getX());
        int colIndex = convertColumnIndexToModel(column); 
        int padding = getIntercellSpacing().height;
        int rowHeight = getRowHeight();
        int rowIndex = (evt.getY() / (padding + rowHeight));

        if (rowIndex < 0 || rowIndex >= getRowCount() || colIndex < 0 )
            return;
        
        doDoubleClick( evt, rowIndex, colIndex );
    }

    /**
     * Set the selected row based on the contents of a column
     * Failure to find a match results in the selection being cleared for this table.
     * @param content the contents to match
     * @param col the column to search in
     * @return success or failure to find a match
     */
    public final boolean setSelectedRow( Object content, int col ) {
        ListSelectionModel lsm = getSelectionModel();
        lsm.setValueIsAdjusting( true );

        if ( content != null ) {
            for ( int i = 0; i < _deluxeModel.getRowCount(); ++i ) {
                if ( _deluxeModel.getValueAt( i, col ).equals( content ) ) {
                    lsm.setSelectionInterval( i, i );
                    lsm.setValueIsAdjusting( true );
                    return true;
                }
            }
        }
        lsm.removeSelectionInterval( lsm.getMinSelectionIndex(), lsm.getMaxSelectionIndex() );
        lsm.setValueIsAdjusting( true );
        return false;
    }

    /**
     * Set the number of visible rows, ala JList
     * @see JList
     * @param rows the number of visible rows, not counting the table header
     */

    public final void setVisibleRowCount( int rows ) {
        Dimension d = getPreferredScrollableViewportSize();

        d.height = ( rows * ( getRowHeight() + getRowMargin() ) + getTableHeader().getSize().height );

        setPreferredScrollableViewportSize( d );
    }

    /**
     * Cause the columns of the tables to automatically size themselves, resizing the table as a whole if necessary.
     *
     * Equivalent to: <code>autoSizeColumns( true, true );</code>
     */
    public final void autoSizeColumns() {
        autoSizeColumns( true, true );
    }

    /**
     * Cause the columns of the table to automatically size themselves.
     * @param allowShrink if true, the table will shrink if the total column widths decreases.  otherwise the table will remain the same size
     * @param allowGrow if true, the table will grow if the total column widths increases.  otherwise the table will remain the same size.
     *
     */

    public final void autoSizeColumns( boolean allowShrink, boolean allowGrow ) {


        Dimension d = getPreferredScrollableViewportSize();
        
        TableColumnModel cm = getColumnModel();
        
        int totwid = 0;

        //
        // loop through the columns, finding the minimum width allowable for each one
        //

        for ( int c = 0; c < getColumnCount(); ++c ) {

            TableColumn col = cm.getColumn( c );

            int modelColumnIndex = col.getModelIndex();

            int maxwid = 0;

            TableCellRenderer renderer = null;

            //
            // first check the width of the header
            //

            renderer = col.getHeaderRenderer();
            if ( renderer != null ) {
                
                maxwid = renderer.getTableCellRendererComponent( this, 
                                                                 col.getHeaderValue(),
                                                                 false, 
                                                                 false, 
                                                                 -1, 
                                                                 0 ).getPreferredSize().width;
                maxwid += 4;
            }

            //
            // now loop through the rows of this column, looking for the widest cell
            //

            for ( int r = 0; r < getRowCount(); ++r ) {
                renderer = getCellRenderer( r, c );
                
                int wid = renderer.getTableCellRendererComponent( this, 
                                                                  getValueAt( r, c ),
                                                                  false, 
                                                                  false, 
                                                                  r, 
                                                                  c ).getPreferredSize().width;
                
                wid += 4;
                maxwid = Math.max( maxwid, wid );
            }
            
            //
            // clamp column width for sanity reasons
            //
            
            maxwid = Math.min( maxwid, _maximumColumnWidth );
            
            //
            // check deluxe table model minimum width for each column
            // 'renderer' should be set to the last row of the column, from above, so no need
            //  to reset it
            //

            if ( renderer != null ) {

                Object minColValue = _deluxeModel.getMinimumWidthColumnValue( modelColumnIndex );
                
                if ( minColValue != null ) {

                    int mincolwid = 
                        renderer.getTableCellRendererComponent( this, 
                                                                minColValue,
                                                                false, 
                                                                false, 
                                                                0, 
                                                                c ).getPreferredSize().width;
                    maxwid = Math.max( mincolwid, maxwid );
                    
                }
            }

            col.setMinWidth( maxwid );
            
            if ( ! _deluxeModel.isColumnExpandable( c ) )
                col.setMaxWidth( maxwid );

            col.setPreferredWidth( maxwid );
            totwid += maxwid;
        }
	

        totwid += cm.getColumnMargin() * ( getColumnCount() + 1 );

        //
        // shrink the last columns if necessary
        //
        
        if ( !allowGrow && totwid > d.width ) {
            int diff = totwid - d.width;
            
            for ( int i = getColumnCount() - 1; i >= 0 && diff > 0 ; --i ) {
                TableColumn col = cm.getColumn( i );
                int thisDiff = Math.min( diff, col.getMinWidth() * 3 / 4 );
                int thisWid = col.getMinWidth() - thisDiff;
                
                col.setMinWidth( thisWid );
                col.setMaxWidth( thisWid );
                col.setPreferredWidth( thisWid );
                diff -= thisDiff;
                totwid -= thisDiff;
            }

        }

        //
        // resize the table only if allowed
        //

        if ( d.width == 0 ||
             ( totwid > d.width && allowGrow ) ||
             ( totwid < d.width && allowShrink ) ) {

            d.width = totwid;
            setPreferredScrollableViewportSize( d );

        }

        sizeColumnsToFit( JTable.AUTO_RESIZE_OFF );

    }

    /**
     * dump the table as a long, formatted string.  Simply calls
     * _deluxeModel.getModelAsString()
     * @see DeluxeTableModel#getModelAsString()
     */
    public synchronized String getTableAsString() {
        return _deluxeModel.getModelAsString();
    }
}
