import javax.swing.*;
import javax.swing.table.*;
import java.util.Vector;

/**
 * A base class that extends the standard AbstractTableModel and provides an additional interface for sorting the tablemodel and obtaining tooltip values for a given cell.
 */
abstract public class DeluxeTableModel extends AbstractTableModel {

    private int _maxSortStates = 2;
    private final Vector _sortStatesVector = new Vector( _maxSortStates );
    
    /**
     * A sort state keeps the sorting order of a column
     */
    public final static class SortState {
        /** The index of the table model column this state refers to */
        public int colNum = -1;
        /** a boolean that indictes whether the column is to be sorted upwards or downwards. up == true, down == false */
        public boolean upDown = false;
        /** SortState contructor */
        public SortState( int col, boolean upDown ) {
            colNum = col;
            this.upDown = upDown;
        }
    }
    
    /**
     * the tablemodel stores a number of previous sort states, so that new model
     * data may be sorted in a similar way to the old data.
     *
     */
    public final void setMaxSortStates( int max ) {
        _maxSortStates = max;
        while ( _sortStatesVector.size() >= _maxSortStates ) {
            _sortStatesVector.removeElementAt( 0 );
        }
        _sortStatesVector.ensureCapacity( _maxSortStates );
    }

    /**
     * should be called by the subclass implementation of sortOnColumn( int ) on each
     * successful sort, so that sorting state may be restored when the table model data
     * is refreshed or replaced
     * @see #sortOnColumn( int )
     */

    protected final void pushSortState( int col, boolean upDown ) {
        while ( _sortStatesVector.size() >= _maxSortStates ) {
            _sortStatesVector.removeElementAt( 0 );
        }
        _sortStatesVector.addElement( new SortState( col, upDown ) );
    }

    /**
     * override this to implement table model sorting
     * @return true for success, false for failure
     * @param col the column index to sort on
     */
    public synchronized boolean sortOnColumn( int col ) {
        return false;
    }

    /**
     * sort a string array, returning an int array containing the new indices.  utility method for use by sortOnColumn( int )
     * @return an int[] containing the new indicies, cannot be null
     * @param arr a String[] containing the strings to be sorted, cannot be null.  The array references are shuffled about, leaving the array in a new, sorted state after method completion.
     * @param upDown a boolean indicating true to sort up, false to sort down.
     * @see #sortOnColumn( int )
     */

    protected static final int[] sortStringArray( String[] arr, boolean upDown ) {
        int[] indices = new int[ arr.length ];
        
        for ( int i = 0; i < arr.length; ++i ) {
            indices[i] = i;
        }
        
        // sort up
        if ( upDown ) {

            for ( int i = 0; i < arr.length; ++i ) {
                for ( int j = i+1; j < arr.length; ++j ) {
                    if ( arr[j].compareTo( arr[i] ) < 0 ) {
                        String tmp = arr[i];
                        arr[i] = arr[j];
                        arr[j] = tmp;
                        int tmp_i = indices[i];
                        indices[i] = indices[j];
                        indices[j] = tmp_i;
                    }
                }
            }
        }
        else {
            for ( int i = 0; i < arr.length; ++i ) {
                for ( int j = i+1; j < arr.length; ++j ) {
                    if ( arr[j].compareTo( arr[i] ) > 0 ) {
                        String tmp = arr[i];
                        arr[i] = arr[j];
                        arr[j] = tmp;
                        int tmp_i = indices[i];
                        indices[i] = indices[j];
                        indices[j] = tmp_i;
                    }
                }
            }
        }

        return indices;
    }

    /**
     * override this to implement tool tips for a cell
     * @return the string for the tooltip, or null for no tooltip
     * @param row the row index of the cell
     * @param col the col index of the cell
     */
    public String getToolTipAt( int row, int col ) {
        return null;
    }

    /**
     * get an object which represents the minimum width of an object value for this column.
     * @return can be null
     * @param col the column index
     */

    public Object getMinimumWidthColumnValue( int col ) {
        return null;
    }
    
    /**
     * can this column expand?  
     * Use this to return true for columns that should remain as small as possible.
     */
    public boolean isColumnExpandable( int col ) {
        return true;
    }
    
    /**
     * convert a data row into a formatted string
     * @param buf the StringBuffer to which to append the row
     * @param values the array of table model values for the row
     * @param sreps the string representations of the values
     * @param widths the the number of spaces wide each column is
     * @param padding a dummy string containing a suitable number of fill characters (typically spaces)
     * @param spaces the literal string to place between each column
     */
    
    private void rowToString( StringBuffer buf, 
                              Object[] values, 
                              String[] sreps, 
                              int[] widths, 
                              String padding, 
                              String spaces ) {
        
        for ( int c = 0; c < values.length; ++c ) {

            //
            // right-align numbers
            //
            
            if ( values[c] instanceof Number ) {
                buf.append( spaces.substring( 0, widths[c] - sreps[c].length() ) );
                buf.append( sreps[c] );
            }
                
            //
            // left-align everything else
            //

            else {

                buf.append( sreps[c] );

                // 
                // fill the field only of this is not the last column
                //
                    
                if ( c < values.length ) {
                    buf.append( spaces.substring( 0, widths[c] - sreps[c].length() ) );
                }
            }
            
            //
            // put in the padding only if it is not the last column
            //

            if ( c < values.length ) {
                buf.append( padding );
            }
        }

        buf.append( "\n" );
    }
    
    /**
     * convert the table model to a long, formatted string.  Includes table headers
     *
     */

    public synchronized String getModelAsString() {

        final int NUM_COLS = getColumnCount();
        final int NUM_ROWS = getRowCount();
        
        int[] widths = new int[ NUM_COLS ]; // the maximum widths for each colummn
       
        // the actual values of the table cells, stored for repeated access
        Object[][] values = new Object[ NUM_ROWS ][ NUM_COLS ];
        String[][] sreps  = new String[ NUM_ROWS ][ NUM_COLS ];
        String[] names    = new String[ NUM_COLS ];
        
        final int PADDING = 2;  // the width of the padding between columns
        final String padding = "  ";
        
        int totalsize = 0;      // the total size of the string needed

        int max_width = 2;      // the maximum width for any row
        
        for ( int c = 0; c < NUM_COLS; ++c ) {
            
            names[c] = getColumnName( c );
            
            widths[c] = names[c].length();
            
            for ( int r = 0; r < NUM_ROWS; ++r ) {
                values[r][c] = getValueAt( r, c );

                sreps[r][c] = values[r][c].toString();

                widths[c] = Math.max( widths[c], sreps[r][c].length() );
                
                totalsize += sreps[r][c].length() + PADDING;
            }
                        
            max_width = Math.max( max_width, widths[c] );
        }
        
        String spaces = new String();
        
        for ( int i = 0; i < max_width; ++i ) {
            spaces += " ";
        }
        
        StringBuffer buf = new StringBuffer( totalsize );

        rowToString( buf, names, names, widths, padding, spaces );
        
        for ( int c = 0; c < NUM_COLS; ++c ) {
            for ( int i = 0; i < widths[c]; ++i ) {
                buf.append( "-" );
            }

            if ( c < NUM_COLS-1 )
                buf.append( padding );
        }
        
        buf.append( "\n" );

        for ( int r = 0; r < NUM_ROWS; ++r ) {
            rowToString( buf, values[r], sreps[r], widths, padding, spaces );
        }
        return buf.toString();
    }
}
