/*
 * View2d.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;

abstract public class View2d extends JPanel implements Map.Listener {
        
        private JMenu     _menu;

        protected AffineTransform _displayTransform = null;
        
        protected Map _map;
        protected ViewControl _vc;
        
        protected JScrollPane _scroll = null;
        
        /** is there transient line that needs to be overwritten? */
        protected boolean transDrawn = false;
        /** the coordinates of the last drawn transient line */
        protected Point[] lastTrans = new Point[] {
                new Point(),
                new Point()
                        };

        protected boolean rubberBandDrawn = false;
        protected boolean transClipDrawn = false;
        protected Rectangle lastRubberBand = new Rectangle();
        protected Rectangle lastTransClip = new Rectangle();
        
        static final protected BasicStroke nilStroke = new BasicStroke( 0 );

        JMenuItem            _zoomInItem      = new JMenuItem( "Zoom In (+)" );
        JMenuItem            _zoomOutItem     = new JMenuItem( "Zoom Out (-)" );
        JMenuItem            _centerItem      = new JMenuItem( "Center View (c-e)" );
        JCheckBoxMenuItem    _showGridItem    = new JCheckBoxMenuItem( "Show Grid (g)", true );
        JRadioButtonMenuItem _gridUnderItem   = new JRadioButtonMenuItem( "Grid Under", true );
        JRadioButtonMenuItem _gridOverItem    = new JRadioButtonMenuItem( "Grid Over", false );
        JCheckBoxMenuItem    _showVertsItem   = new JCheckBoxMenuItem( "Show Verts", false );
        JCheckBoxMenuItem    _showLinesItem   = new JCheckBoxMenuItem( "Show Lines", true );
        JCheckBoxMenuItem    _showLineArrowsItem = new JCheckBoxMenuItem( "Show Line Arrows", false );
        JCheckBoxMenuItem    _showSidesItem   = new JCheckBoxMenuItem( "Show Sides", false );
        JCheckBoxMenuItem    _fillSectItem    = new JCheckBoxMenuItem( "Fill Sectors", true );
        JCheckBoxMenuItem    _fillSectAlphaItem   = new JCheckBoxMenuItem( "Fill Sector Alpha", true );
        
        Dimension _panelSize = new Dimension();
        
        Controller2d _controller = null;
        
        protected Color _bgColor = new Color( 0.07f, 0.15f, 0.10f );
        
        /**
         * how many pixels represent one map unit
         */
        protected float _zoom = 0.5f;
        
        public float getZoom() {
                return _zoom;
        }

        protected static final String ZOOM_IN_CMD   = "ZoomIn";
        protected static final String ZOOM_OUT_CMD  = "ZoomOut";
        protected static final String SHOW_GRID_CMD = "ShowGrid";
        protected static final String CENTER_CMD    = "Center";
        
        ActionListener menuListener = new ActionListener() {
                        public void actionPerformed( ActionEvent evt ) {
                                Object source = evt.getSource();
                                String command = evt.getActionCommand();
                                
                                if ( source == _showGridItem ) {
                                        _gridUnderItem.setEnabled( _showGridItem.isSelected() );
                                        _gridOverItem.setEnabled( _showGridItem.isSelected() );
                                }
                                else if ( source == _fillSectItem ) {
                                        _fillSectAlphaItem.setEnabled( _fillSectItem.isSelected() );
                                }
                                else if ( source == _zoomInItem || command.equals( ZOOM_IN_CMD ) ) {
                                        modifyZoom( 0.1f );
                                }
                                else if ( source == _zoomOutItem || command.equals( ZOOM_OUT_CMD ) ) {
                                        modifyZoom( - 0.1f );
                                }
                                else if ( command.equals( SHOW_GRID_CMD ) ) {
                                        _showGridItem.doClick();
                                }
                                else if ( source == _centerItem || command.equals( CENTER_CMD ) ) {
                                        
                                        centerView( true, true );
                                }
                                
                                repaint();
                        }
                };
        
        /**
         * center the view
         * @param centerOnSelection true if the selection(s) are used for centering
         * @param multiSelection true if the center point is the average of the selections
         *
         */
        abstract public void centerView( boolean centerOnSelection, boolean multiSelection );

        protected void modifyZoom( float mod ) {
                setZoom( _zoom + mod );
        }
        
        protected void setZoom( float zoom ) {
                
                Point ocp = null;
                Dimension ov = null;
                
                if ( _scroll != null ) {
                        Point vp = _scroll.getViewport().getViewPosition();
                        Dimension extent = _scroll.getViewport().getExtentSize();
                        
                        ocp = new Point( vp.x + extent.width / 2,
                                         vp.y + extent.height / 2 );
                        ov = new Dimension( _panelSize );
                }

                
                _zoom = Math.max( 0.1f, zoom );
                _zoom = Math.min( 2, _zoom );
                
                _vc.setInfoZoom( _zoom );

                updatePanelSize();

                if ( ocp != null ) {
                        Dimension extent = _scroll.getViewport().getExtentSize();
                        Point cp = new Point( ocp.x * _panelSize.width / ov.width - extent.width / 2,
                                              ocp.y * _panelSize.height / ov.height - extent.height / 2 );
                        cp.x = Math.min( _panelSize.width - extent.width, cp.x );
                        cp.x = Math.max( 0, cp.x );
                        
                        cp.y = Math.min( _panelSize.height - extent.height, cp.y );
                        cp.y = Math.max( 0, cp.y );

                        _scroll.getViewport().setViewPosition( cp );
                }

                
        }

        /**
         * overridden to return the dimensions of the 2-d slice of the
         * map which we are displaying
         */

        abstract public Dimension getMapSliceSize();
        
        /**
         * this utilizes the protected method getMapSliceSize to determine the appropriate panel size
         * @see #getMapSliceSize()
         */
        protected void updatePanelSize() {

                Dimension d = getMapSliceSize();
                
                _displayTransform = new AffineTransform( 1f, 0f, 0f, -1f, d.width / 2, d.height / 2 );

                _panelSize.setSize( (int) ( _zoom * d.width ),
                                    (int) ( _zoom * d.height ) );
                
                setSize(          _panelSize );
                setPreferredSize( _panelSize );
                setMinimumSize(   _panelSize );
                setMaximumSize(   _panelSize );

                _horizontalRule.setPreferredWidth( _panelSize.width );
                _verticalRule.setPreferredHeight( _panelSize.height );

                // TODO: fix this ugly repaint hack

                _horizontalRule.invalidate();
                _horizontalRule.revalidate();
                _verticalRule.invalidate();
                _verticalRule.revalidate();

                invalidate();
                revalidate();
        }

        void setController( ViewControl vc, Controller2d controller ) {
                _vc = vc;
                _controller = controller;
        }

        public View2d(  JMenu menu ) {
                _menu = menu;

                _menu.removeAll();

                ButtonGroup bg = new ButtonGroup();

                bg.add( _gridUnderItem );
                bg.add( _gridOverItem );

                _zoomInItem.addActionListener( menuListener );
                _zoomOutItem.addActionListener( menuListener );
                _centerItem.addActionListener( menuListener );
                
                registerKeyboardAction( menuListener,
                                        CENTER_CMD,
                                        KeyStroke.getKeyStroke( KeyEvent.VK_E, Event.CTRL_MASK ),
                                        JComponent.WHEN_FOCUSED );

                registerKeyboardAction( menuListener,
                                        ZOOM_IN_CMD,
                                        KeyStroke.getKeyStroke( KeyEvent.VK_EQUALS, 0 ),
                                        JComponent.WHEN_FOCUSED );

                registerKeyboardAction( menuListener,
                                        ZOOM_OUT_CMD,
                                        KeyStroke.getKeyStroke( KeyEvent.VK_MINUS, 0 ),
                                        JComponent.WHEN_FOCUSED );

                registerKeyboardAction( menuListener,
                                        SHOW_GRID_CMD,
                                        KeyStroke.getKeyStroke( KeyEvent.VK_G, 0 ),
                                        JComponent.WHEN_FOCUSED );
                
                _showGridItem.addActionListener( menuListener );
                _gridUnderItem.addActionListener( menuListener );
                _gridOverItem.addActionListener( menuListener );

                _showVertsItem.addActionListener( menuListener );
                _showLinesItem.addActionListener( menuListener );
                _showLineArrowsItem.addActionListener( menuListener );
                _showSidesItem.addActionListener( menuListener );
                _fillSectItem.addActionListener( menuListener );
                _fillSectAlphaItem.addActionListener( menuListener );
                
                _menu.add( _zoomInItem );
                _menu.add( _zoomOutItem );
                _menu.add( _centerItem );

                _menu.addSeparator();

                _menu.add( _showGridItem );

                _menu.addSeparator();

                _menu.add( _gridUnderItem );
                _menu.add( _gridOverItem );

                _menu.addSeparator();

                _menu.add( _showVertsItem );
                _menu.add( _showLinesItem );
                _menu.add( _showLineArrowsItem );
                _menu.add( _showSidesItem );
                _menu.add( _fillSectItem );
                _menu.add( _fillSectAlphaItem );

                setBackground( _bgColor );//new Color( 0.1f, 0.1f, 0.1f ) );                
        }

        void setMap( Map map ) {
                if ( _map != null )
                        _map.listeners.removeElement( this );
                
                _map = map;
                
                if ( _map != null )
                        _map.listeners.addElement( this );
                
                updatePanelSize();
                
                centerView( false, false );

                repaint();
        }
        
        protected Color NEW_LINE_COLOR              = Color.red;
        protected Color SIDE_COLOR                  = new Color( 0.5f, 0.8f, 1.0f );
        protected Color NEW_SECT_DEFAULT_LINE_COLOR = Color.pink;
        protected Color LINE_COLOR                  = Color.white;
        protected Color SPEC_LINE_COLOR             = new Color( 0.1f, 1.0f, 0.3f );
        protected Color SPEC_LINE_SELECTED_COLOR    = new Color( 1.0f, 0.8f, 0.0f );

        protected Color SECT_FILL                   = new Color( 0.3f, 0.3f, 0.3f );
        protected Color SECT_FILL_ALPHA             = new Color( 0.5f, 0.5f, 0.5f, 0.5f );
        protected Color SELECTED_SECT_FILL          = new Color( 0.7f, 0.7f, 0.5f );
        protected Color SELECTED_SECT_FILL_ALPHA    = new Color( 0.7f, 0.7f, 0.5f, 0.5f );

        protected Color SPEC_SECT_FILL                   = new Color( 0.5f, 0.3f, 0.5f );
        protected Color SPEC_SECT_FILL_ALPHA             = new Color( 0.5f, 0.3f, 0.5f, 0.5f );
        protected Color SELECTED_SPEC_SECT_FILL          = new Color( 0.7f, 0.5f, 0.6f );
        protected Color SELECTED_SPEC_SECT_FILL_ALPHA    = new Color( 0.7f, 0.5f, 0.6f, 0.5f );

        protected Color LIGHT_SOURCE_COLOR   = new Color( 0.9f, 0.9f, 1f );

        protected Color VERTS_COLOR          = Color.blue;
        
        protected Color SELECTED_COLOR      = Color.orange;
        
        protected Color GRID_COLOR_MAJOR = new Color( 0.40f, 0.40f, 0.40f );
        protected Color GRID_COLOR       = GRID_COLOR_MAJOR.darker();
        
        protected Color RUBBER_COLOR     = Color.cyan;

        int grid_x_start, grid_y_start, grid_x_end, grid_y_end;
        
        /**
         * drawGrid should be useful for any orthogonal 2d view
         */
        
        void drawGrid( Graphics g ) {
                
                Rectangle clip = g.getClipBounds();
                
                g.setColor( GRID_COLOR );
                
                int gridSize = _vc.getGridSize();
                
                grid_x_start = ( clip.x / gridSize - 1 ) * gridSize;
                grid_y_start = ( clip.y / gridSize - 1 ) * gridSize;
                grid_x_end   = grid_x_start + clip.width + gridSize * 2;
                grid_y_end   = grid_y_start + clip.height + gridSize * 2;
                
                for ( int i = grid_x_start; i < grid_x_end; i += gridSize ) {
                        g.drawLine( i, grid_y_start, i, grid_y_end );
                }
                
                for ( int i = grid_y_start; i < grid_y_end; i += gridSize ) {
                        g.drawLine( grid_x_start, i, grid_x_end, i );
                }

                g.setColor( Color.white );
                if ( grid_x_start < 0 && grid_x_end > 0 )
                        g.drawLine( 0, grid_y_start, 0, grid_y_end );
                if ( grid_y_start < 0 && grid_y_end > 0 )
                        g.drawLine( grid_x_start, 0, grid_x_end, 0 );
        }
        
        /**
         * overriden to handle transient drawing.  rubber banding is handled by this class
         */
        void drawTransient( Graphics g ) {
                    
                if ( rubberBandDrawn ) {
                        g.setXORMode( RUBBER_COLOR );
                        g.drawRect( lastRubberBand.x, 
                                    lastRubberBand.y, 
                                    lastRubberBand.width, 
                                    lastRubberBand.height );
                        rubberBandDrawn = false;
                        g.setPaintMode();
                }

      
                if ( transClipDrawn ) {
                        g.setXORMode( RUBBER_COLOR );
                        g.drawRect( lastTransClip.x, 
                                    lastTransClip.y, 
                                    lastTransClip.width, 
                                    lastTransClip.height );
                        transClipDrawn = false;
                        g.setPaintMode();
                }
        }
        
        public void mapChanged( Rectangle clip ) {
                repaint();
        }
        
        public void mapTransientChanged() {
                Graphics2D g = (Graphics2D) getGraphics();

                if ( g == null )
                        return;

                g.setClip( getVisibleRect() );
                g.scale( _zoom, _zoom );
                g.transform( _displayTransform );
                drawTransient( g );
        }

        public void mapSizeChanged() {
                updatePanelSize();
        }

        public static final int HORIZONTAL = 0;
        public static final int VERTICAL = 1;

        public JComponent getRule( int type ) {
                if ( type == HORIZONTAL )
                        return _horizontalRule;
                else
                        return _verticalRule;
        }
        
        protected Rule _horizontalRule = new Rule( HORIZONTAL );
        protected Rule _verticalRule   = new Rule( VERTICAL );
        
        protected class Rule extends JComponent {
                public static final int SIZE = 24;
                public static final int HALF_SIZE = 12;
                public static final int QUARTER_SIZE = 6;

                private Dimension _prefSize = new Dimension();

                public int orientation;
                private int increment;
                private int units;

                private Color _color = new Color( 250, 230, 150 );

                public Rule( int o ) {
                        orientation = o;
                        setIncrementAndUnits();

                        if ( orientation == HORIZONTAL )
                                _prefSize.height = SIZE;
                        else
                                _prefSize.width = SIZE;
                }

                private void setIncrementAndUnits() {
                        increment = units / 2;
                }

                public int getIncrement() {
                        return increment;
                }

                public Dimension getPreferredSize() {
                        return _prefSize;
                }

                void setPreferredWidth( int w ) {
                        _prefSize.width = w;
                }

                void setPreferredHeight( int h ) {
                        _prefSize.height = h;
                }

                private Font _theFont = new Font( "SansSerif", Font.PLAIN, 10 );

                private int map_half_x, map_half_y;
                
                public void paintComponent( Graphics g ) {

                        super.paintComponent( g );

                        Rectangle clip = g.getClipBounds();

                        g.setColor( _color );
                        
                        if ( orientation == HORIZONTAL )
                                clip.width = Math.min( _prefSize.width - clip.x, clip.width );
                        else
                                clip.height = Math.min( _prefSize.height - clip.y, clip.height );
                        
                        // Fill clipping area with dirty brown/orange.
                        g.fillRect( clip.x, clip.y, clip.width, clip.height );

                        // Do the ruler labels in a small font that's black.
                        g.setFont( _theFont ); 
                        g.setColor( Color.black );
                        
                        // Some vars we need.
                        int gridSize = _vc.getGridSize();
                        float my_increment = gridSize * _zoom;
                        Dimension mapSize = _map.getMapSize();
                        
                        if ( orientation == HORIZONTAL ) {

                                map_half_x = mapSize.width / 2;

                                int startpos = (int) ( (float) clip.x / my_increment );
                                startpos = ( startpos / 2 ) * 2;
                                int endpos   = (int) ( startpos + ( (float) clip.width / my_increment ) );
                                
                                for ( int i = startpos; i <= endpos; i += 2 ) {
                                        int x = (int) ( (float) i * my_increment );
                                        
                                        g.setColor( Color.black );
                                        g.drawLine( x, SIZE, x, QUARTER_SIZE );
                                        g.setColor( Color.blue );
                                        g.drawString( "" + ( ( i * gridSize ) - map_half_x ), x-10, HALF_SIZE );
                                }
                                        
                                return;
                        }

                        else {
                                map_half_y = mapSize.height / 2;
                                int startpos = (int) ( (float) clip.y / my_increment );
                                startpos = ( startpos / 2 ) * 2;
                                int endpos   = (int) ( startpos + ( (float) clip.height / my_increment ) );
                                
                                for ( int i = startpos; i <= endpos; i += 2 ) {
                                        int y = (int) ( (float) i * my_increment );
                                        
                                        g.setColor( Color.black );
                                        g.drawLine( SIZE, y, QUARTER_SIZE, y );
                                        g.setColor( Color.blue );
                                        g.drawString( "" + ( map_half_y - ( i * gridSize ) ), 0, y );
                                }
                                        
                                return;
                        }
                }
        }

        public final JScrollPane getScrolledView() {
                if ( _scroll == null ) {

                        _scroll = new JScrollPane( this );
                        
                        JScrollBar sbar = null;

                        sbar = _scroll.getVerticalScrollBar();
                        if ( sbar != null ) {
                                sbar.setUnitIncrement( 32 );
                                sbar.setBlockIncrement( 128 );
                        }

                        sbar = _scroll.getHorizontalScrollBar();
                        if ( sbar != null ) {
                                sbar.setUnitIncrement( 32 );
                                sbar.setBlockIncrement( 128 );
                        }
                        
                        _scroll.setColumnHeaderView( _horizontalRule );
                        _scroll.setRowHeaderView( _verticalRule );
                        
                        setAutoscrolls( true );
                }

                return _scroll;
        }

        public JScrollPane getScrollPane() {
                return _scroll;
        }

}
